from typing import TypedDict


class KeepJobs(TypedDict, total=False):
    """
    Specify which jobs to keep after finishing. If both age and count are
    specified, then the jobs kept will be the ones that satisfies both
    properties.
    """

    age: int
    """
    Maximum age in seconds for job to be kept.
    """

    count: int
    """
    Maximum count of jobs to be kept.
    """



import asyncio
from bullmq.redis_connection import RedisConnection
from bullmq.types import QueueBaseOptions, RetryJobsOptions, JobOptions, PromoteJobsOptions
from bullmq.utils import extract_result
from bullmq.scripts import Scripts
from bullmq.job import Job


class Queue:
    """
    Instantiate a Queue object
    """

    def __init__(self, name: str, opts: QueueBaseOptions = {}):
        """
        Initialize a connection
        """
        self.name = name
        redisOpts = opts.get("connection", {})
        self.redisConnection = RedisConnection(redisOpts)
        self.client = self.redisConnection.conn
        self.opts = opts
        self.jobsOpts = opts.get("defaultJobOptions", {})
        self.prefix = opts.get("prefix", "bull")
        self.scripts = Scripts(
            self.prefix, name, self.redisConnection)
        self.keys = self.scripts.queue_keys.getKeys(name)
        self.qualifiedName = self.scripts.queue_keys.getQueueQualifiedName(name)

    def toKey(self, type: str):
        return self.scripts.queue_keys.toKey(self.name, type)

    async def add(self, name: str, data, opts: JobOptions = {}):
        """
        Adds a new job to the queue.

        @param name: Name of the job to be added to the queue,.
        @param data: Arbitrary data to append to the job.
        @param opts: Job options that affects how the job is going to be processed.
        """
        job = Job(self, name, data, opts)
        job_id = await self.scripts.addJob(job)
        job.id = job_id
        return job

    async def addBulk(self, jobs: list[dict[str,dict | str]]):
        """
        Adds an array of jobs to the queue. This method may be faster than adding
        one job at a time in a sequence
        """
        jobs_data = []
        for job in jobs:
            opts = {}
            opts.update(self.jobsOpts)
            opts.update(job.get("opts", {}))
            
            jobs_data.append({
                "name": job.get("name"),
                "data": job.get("data"),
                "opts": opts
            })

        result = []
        async with self.redisConnection.conn.pipeline(transaction=True) as pipe:
            for job_data in jobs_data:
                current_job_opts = job_data.get("opts", {})
                job = Job(
                    queue=self,
                    name=job_data.get("name"),
                    data=job_data.get("data"),
                    opts=current_job_opts,
                    job_id = current_job_opts.get("jobId")
                    )
                job_id = await self.scripts.addJob(job, pipe)
                job.id = job_id
                result.append(job)
            job_ids = await pipe.execute()
            for index, job_id in enumerate(job_ids):
                result[index].id = job_id

        return result

    def pause(self):
        """
        Pauses the processing of this queue globally.

        We use an atomic RENAME operation on the wait queue. Since
        we have blocking calls with BRPOPLPUSH on the wait queue, as long as the queue
        is renamed to 'paused', no new jobs will be processed (the current ones
        will run until finalized).

        Adding jobs requires a LUA script to check first if the paused list exist
        and in that case it will add it there instead of the wait list.
        """
        return self.scripts.pause(True)

    def resume(self):
        """
        Resumes the processing of this queue globally.

        The method reverses the pause operation by resuming the processing of the
        queue.
        """
        return self.scripts.pause(False)

    async def isPaused(self):
        """
        Returns true if the queue is currently paused.
        """
        paused_key_exists = await self.client.hexists(self.keys["meta"], "paused")
        return paused_key_exists == 1

    def getRateLimitTtl(self):
        """
        Returns the time to live for a rate limited key in milliseconds.
        """
        return self.client.pttl(self.keys["limiter"])

    async def obliterate(self, force: bool = False):
        """
        Completely destroys the queue and all of its contents irreversibly.
        This method will the *pause* the queue and requires that there are no
        active jobs. It is possible to bypass this requirement, i.e. not
        having active jobs using the "force" option.

        Note: This operation requires to iterate on all the jobs stored in the queue
        and can be slow for very large queues.

        @param opts: Obliterate options.
        """
        await self.pause()
        while True:
            cursor = await self.scripts.obliterate(1000, force)
            if cursor is None or cursor == 0 or cursor == "0":
                break

    async def retryJobs(self, opts: RetryJobsOptions = {}):
        """
        Retry all the failed or completed jobs.
        """
        while True:
            cursor = await self.scripts.retryJobs(
                opts.get("state"),
                opts.get("count"),
                opts.get("timestamp")
            )
            if cursor is None or cursor == 0 or cursor == "0":
                break

    async def promoteJobs(self, opts: PromoteJobsOptions = {}):
        """
        Retry all the delayed jobs.
        """
        while True:
            cursor = await self.scripts.promoteJobs(
                opts.get("count")
            )
            if cursor is None or cursor == 0 or cursor == "0":
                break

    def trimEvents(self, maxLength: int):
        """
        Trim the event stream to an approximately maxLength.

        @param maxLength:
        """
        return self.client.xtrim(self.keys["events"], maxlen = maxLength, approximate = "~")

    def removeDeprecatedPriorityKey(self):
        """
        Delete old priority helper key.
        """
        return self.client.delete(self.toKey("priority"))

    async def getJobCountByTypes(self, *types):
      result = await self.getJobCounts(*types)
      sum = 0
      for attribute in result:
        sum += result[attribute]
      return sum

    async def getJobCounts(self, *types):
        """
        Returns the job counts for each type specified or every list/set in the queue by default.

        @returns: An object, key (type) and value (count)
        """
        current_types = self.sanitizeJobTypes(types)

        responses = await self.scripts.getCounts(current_types)
        counts = {}

        for index, val in enumerate(responses):
            counts[current_types[index]] = val or 0
        return counts

    async def clean(self, grace: int, limit: int, type: str):
        """
        Cleans jobs from a queue. Similar to drain but keeps jobs within a certain
        grace period
        
        * @returns: Id jobs from the deleted records
        """
        jobs = await self.scripts.cleanJobsInSet(type, grace, limit)

        return jobs

    def getJobState(self, job_id: str):
        return self.scripts.getState(job_id)

    def getCompletedCount(self):
        return self.getJobCountByTypes('completed')

    def getFailedCount(self):
        return self.getJobCountByTypes('failed')

    def getActive(self, start = 0, end=-1):
        return self.getJobs(['active'], start, end, True)

    def getCompleted(self, start = 0, end=-1):
        return self.getJobs(['completed'], start, end, False)

    def getDelayed(self, start = 0, end=-1):
        return self.getJobs(['delayed'], start, end, True)

    def getFailed(self, start = 0, end=-1):
        return self.getJobs(['completed'], start, end, False)

    def getPrioritized(self, start = 0, end=-1):
        return self.getJobs(['prioritized'], start, end, True)

    def getWaiting(self, start = 0, end=-1):
        return self.getJobs(['waiting'], start, end, True)

    def getWaitingChildren(self, start = 0, end=-1):
        return self.getJobs(['waiting-children'], start, end, True)

    async def getJobs(self, types, start=0, end=-1, asc:bool=False):
        current_types = self.sanitizeJobTypes(types)
        job_ids = await self.scripts.getRanges(current_types, start, end, asc)
        tasks = [asyncio.create_task(Job.fromId(self, i)) for i in job_ids]   
        job_set, _ = await asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)
        jobs = [extract_result(job_task) for job_task in job_set]
        jobs_len = len(jobs)

        # we filter `None` out to remove:
        jobs = list(filter(lambda j: j is not None, jobs))

        for index, job_id in enumerate(job_ids):
            pivot_job = jobs[index]

            for i in range(index,jobs_len):
                current_job = jobs[i]
                if current_job and current_job.id == job_id:
                    jobs[index] = current_job
                    jobs[i] = pivot_job
                    continue

        return jobs

    def sanitizeJobTypes(self, types):
        current_types = list(types)

        if len(types) > 0:
            sanitized_types = current_types.copy()

            try:
                sanitized_types.index('waiting')
                sanitized_types.append('paused')
            except ValueError:
                pass
            set_res = set(sanitized_types)
            list_res = (list(set_res))

            return list_res
        return [
            'active',
            'completed',
            'delayed',
            'failed',
            'paused',
            'waiting',
            'waiting-children'
        ]

    def close(self):
        """
        Close the queue instance.
        """
        return self.redisConnection.close()

    def remove(self, job_id: str, opts: dict = {}):
        return self.scripts.remove(job_id, opts.get("removeChildren", True))



from bullmq.types.backoff_options import BackoffOptions
from bullmq.types.keep_jobs import KeepJobs
from bullmq.types.job_options import JobOptions
from bullmq.types.promote_jobs_options import PromoteJobsOptions
from bullmq.types.queue_options import QueueBaseOptions
from bullmq.types.worker_options import WorkerOptions
from bullmq.types.retry_jobs_options import RetryJobsOptions




from typing import TypedDict, Any


class QueueBaseOptions(TypedDict, total=False):
    """
    Options for the Queue class.
    """

    prefix: str
    connection: dict[str, Any]
    """
    Prefix for all queue keys.
    """




"""
    This class is used to load and execute Lua scripts.
    It is a wrapper around the Redis client.
"""

from __future__ import annotations
from redis import Redis
from bullmq.queue_keys import QueueKeys
from bullmq.error_code import ErrorCode
from bullmq.utils import isRedisVersionLowerThan, get_parent_key
from typing import Any, TYPE_CHECKING
if TYPE_CHECKING:
    from bullmq.job import Job
    from bullmq.redis_connection import RedisConnection

import time
import json
import msgpack
import os


basePath = os.path.dirname(os.path.realpath(__file__))


class Scripts:

    def __init__(self, prefix: str, queueName: str, redisConnection: RedisConnection):
        self.prefix = prefix
        self.queueName = queueName
        self.keys = {}
        self.redisConnection = redisConnection
        self.redisClient = redisConnection.conn
        self.commands = {
            "addStandardJob": self.redisClient.register_script(self.getScript("addStandardJob-7.lua")), 
            "addDelayedJob": self.redisClient.register_script(self.getScript("addDelayedJob-6.lua")), 
            "addParentJob": self.redisClient.register_script(self.getScript("addParentJob-4.lua")),
            "addPrioritizedJob": self.redisClient.register_script(self.getScript("addPrioritizedJob-7.lua")),
            "changePriority": self.redisClient.register_script(self.getScript("changePriority-6.lua")),
            "cleanJobsInSet": self.redisClient.register_script(self.getScript("cleanJobsInSet-2.lua")),
            "extendLock": self.redisClient.register_script(self.getScript("extendLock-2.lua")),
            "getCounts": self.redisClient.register_script(self.getScript("getCounts-1.lua")), #
            "getRanges": self.redisClient.register_script(self.getScript("getRanges-1.lua")), #
            "getState": self.redisClient.register_script(self.getScript("getState-8.lua")),
            "getStateV2": self.redisClient.register_script(self.getScript("getStateV2-8.lua")),
            "isJobInList": self.redisClient.register_script(self.getScript("isJobInList-1.lua")),
            "moveStalledJobsToWait": self.redisClient.register_script(self.getScript("moveStalledJobsToWait-9.lua")),
            "moveToActive": self.redisClient.register_script(self.getScript("moveToActive-11.lua")),
            "moveToDelayed": self.redisClient.register_script(self.getScript("moveToDelayed-7.lua")),
            "moveToFinished": self.redisClient.register_script(self.getScript("moveToFinished-14.lua")),
            "moveToWaitingChildren": self.redisClient.register_script(self.getScript("moveToWaitingChildren-4.lua")),
            "obliterate": self.redisClient.register_script(self.getScript("obliterate-2.lua")),
            "pause": self.redisClient.register_script(self.getScript("pause-7.lua")),
            "promote": self.redisClient.register_script(self.getScript("promote-8.lua")),
            "removeJob": self.redisClient.register_script(self.getScript("removeJob-1.lua")),
            "reprocessJob": self.redisClient.register_script(self.getScript("reprocessJob-7.lua")),
            "retryJob": self.redisClient.register_script(self.getScript("retryJob-10.lua")),
            "moveJobsToWait": self.redisClient.register_script(self.getScript("moveJobsToWait-7.lua")),
            "saveStacktrace": self.redisClient.register_script(self.getScript("saveStacktrace-1.lua")),
            "updateData": self.redisClient.register_script(self.getScript("updateData-1.lua")),
            "updateProgress": self.redisClient.register_script(self.getScript("updateProgress-3.lua")),
        }

        self.queue_keys = QueueKeys(prefix)
        self.keys = self.queue_keys.getKeys(queueName)

    def resetQueueKeys(self, queue_name: str):
        self.queueName = queue_name
        self.keys = self.queue_keys.getKeys(queue_name)

    def toKey(self, name: str):
        return self.queue_keys.toKey(self.queueName, name)

    def getScript(self, name: str):
        """
        Get a script by name
        """
        file = open(f"{basePath}/commands/{name}", "r")
        data = file.read()
        file.close()
        return data

    def getKeys(self, keys: list[str]):
        def mapKey(key):
            return self.keys[key]
        return list(map(mapKey, keys))

    def addJobArgs(self, job: Job, waiting_children_key: str|None):
        #  We are still lacking some arguments here:
        #  ARGV[1] msgpacked arguments array
        #         [9]  repeat job key

        jsonData = json.dumps(job.data, separators=(',', ':'))
        packedOpts = msgpack.packb(job.opts)

        parent = job.parent
        parentKey = job.parentKey

        packedArgs = msgpack.packb(
            [self.keys[""], job.id or "", job.name, job.timestamp, job.parentKey,
                waiting_children_key,
                f"{parentKey}:dependencies" if parentKey else None, parent],use_bin_type=True)
        
        return [packedArgs, jsonData, packedOpts]

    def addJob(self, job: Job, pipe = None):
        """
        Add an item to the queue
        """
        if job.opts.get("delay"):
            return self.addDelayedJob(job, job.opts.get("delay"), pipe)
        elif job.opts.get("priority"):
            return self.addPrioritizedJob(job, job.opts.get("priority"), pipe)
        else:
            return self.addStandardJob(job, job.timestamp, pipe)

    def addStandardJob(self, job: Job, timestamp: int, pipe = None):
        """
        Add a standard job to the queue
        """
        keys = self.getKeys(['wait', 'paused', 'meta', 'id',
                             'completed', 'events', 'marker'])
        args = self.addJobArgs(job, None)
        args.append(timestamp)

        return self.commands["addStandardJob"](keys=keys, args=args, client=pipe)
    
    def addDelayedJob(self, job: Job, timestamp: int, pipe = None):
        """
        Add a delayed job to the queue
        """
        keys = self.getKeys(['marker', 'meta', 'id',
                            'delayed', 'completed', 'events'])
        args = self.addJobArgs(job, None)
        args.append(timestamp)

        return self.commands["addDelayedJob"](keys=keys, args=args, client=pipe)
    
    def addPrioritizedJob(self, job: Job, timestamp: int, pipe = None):
        """
        Add a prioritized job to the queue
        """
        keys = self.getKeys(['marker', 'meta', 'id',
                            'prioritized', 'completed', 'events', 'pc'])
        args = self.addJobArgs(job, None)
        args.append(timestamp)

        return self.commands["addPrioritizedJob"](keys=keys, args=args, client=pipe)

    def addParentJob(self, job: Job, waiting_children_key: str, pipe = None):
        """
        Add a job to the queue that is a parent
        """
        keys = self.getKeys(['meta', 'id', 'completed', 'events'])
        
        args = self.addJobArgs(job, waiting_children_key)

        return self.commands["addParentJob"](keys=keys, args=args, client=pipe)

    def cleanJobsInSetArgs(self, set: str, grace: int, limit:int = 0):
        keys = [self.toKey(set),
                self.keys['events']]
        args = [self.keys[''], round(time.time() * 1000) - grace, limit, set]

        return (keys, args)

    def cleanJobsInSet(self, set: str, grace: int = 0, limit:int = 0):
        keys, args = self.cleanJobsInSetArgs(set, grace, limit)
        return self.commands["cleanJobsInSet"](keys=keys, args=args)

    def moveToWaitingChildrenArgs(self, job_id, token, opts: dict = {}):
        keys = [self.toKey(job_id) + ":lock",
                self.keys['active'],
                self.keys['waiting-children'],
                self.toKey(job_id)]
        child_key = opts.get("child") if opts else None
        args = [token, get_parent_key(child_key) or "", round(time.time() * 1000), job_id,
                "1" if opts.get("skipAttempt") else "0"]

        return (keys, args)

    async def moveToWaitingChildren(self, job_id, token, opts):
        keys, args = self.moveToWaitingChildrenArgs(job_id, token, opts)
        result = await self.commands["moveToWaitingChildren"](keys=keys, args=args)

        if result is not None:
            if result == 1:
                return False
            elif result == 0:
                return True
            elif result < 0:
                raise self.finishedErrors(result, job_id, 'moveToWaitingChildren', 'active')
        return None

    def getRangesArgs(self, types, start: int = 0, end: int = 1, asc: bool = False):
        transformed_types = []
        for type in types:
            transformed_types.append("wait" if type == "waiting" else type)

        keys = self.getKeys([''])
        args = [start, end, "1" if asc else "0"] + transformed_types

        return (keys, args)

    async def getRanges(self, types, start: int = 0, end: int = 1, asc: bool = False):
        commands = []

        switcher = {
            "completed": "zrange",
            "delayed": "zrange",
            "failed": "zrange",
            "priority": "zrange",
            "repeat": "zrange",
            "waiting-children": "zrange",
            "active": "lrange",
            "paused": "lrange",
            "wait": "lrange"
        }
        transformed_types = []
        for type in types:
            transformed_type = "wait" if type == "waiting" else type
            transformed_types.append(transformed_type)
            commands.append(switcher.get(transformed_type))

        keys, args = self.getRangesArgs(transformed_types, start, end, asc)

        responses = await self.commands["getRanges"](keys=keys, args=args)

        results = []
        for i, response in enumerate(responses):
            result = response or []

            if asc and commands[i] == "lrange":
                results+=result.reverse()
            else:
                results+=result

        return results

    def saveStacktraceArgs(self, job_id: str, stacktrace: str, failedReason: str):
        keys = [self.toKey(job_id)]
        args = [stacktrace, failedReason]

        return (keys, args)

    def retryJobArgs(self, job_id: str, lifo: bool, token: str, opts: dict = {}):
        keys = self.getKeys(['active', 'wait', 'paused'])
        keys.append(self.toKey(job_id))
        keys.append(self.keys['meta'])
        keys.append(self.keys['events'])
        keys.append(self.keys['delayed'])
        keys.append(self.keys['prioritized'])
        keys.append(self.keys['pc'])
        keys.append(self.keys['marker'])

        push_cmd = "RPUSH" if lifo else "LPUSH"

        args = [self.keys[''], round(time.time() * 1000), push_cmd,
            job_id, token, "1" if opts.get("skipAttempt") else "0"]

        return (keys, args)

    def moveToDelayedArgs(self, job_id: str, timestamp: int, token: str, delay: int = 0, opts: dict = {}):
        max_timestamp = max(0, timestamp or 0)

        if timestamp > 0:
            max_timestamp = max_timestamp * 0x1000 + (convert_to_int(job_id) & 0xfff)

        keys = self.getKeys(['marker', 'active', 'prioritized', 'delayed'])
        keys.append(self.toKey(job_id))
        keys.append(self.keys['events'])
        keys.append(self.keys['meta'])

        args = [self.keys[''], round(time.time() * 1000), str(max_timestamp),
            job_id, token, delay, "1" if opts.get("skipAttempt") else "0" ]

        return (keys, args)

    async def moveToDelayed(self, job_id: str, timestamp: int, delay: int, token: str = "0"):
        keys, args = self.moveToDelayedArgs(job_id, timestamp, token, delay)

        result = await self.commands["moveToDelayed"](keys=keys, args=args)

        if result is not None:
            if result < 0:
                raise self.finishedErrors(result, job_id, 'moveToDelayed', 'active')
        return None

    def promoteArgs(self, job_id: str):
        keys = self.getKeys(['delayed', 'wait', 'paused', 'meta', 'prioritized', 'pc', 'events', 'marker'])
        keys.append(self.toKey(job_id))
        keys.append(self.keys['events'])
        keys.append(self.keys['paused'])
        keys.append(self.keys['meta'])

        args = [self.keys[''], job_id]

        return (keys, args)

    async def promote(self, job_id: str):
        keys, args = self.promoteArgs(job_id)

        result = await self.commands["promote"](keys=keys, args=args)

        if result is not None:
            if result < 0:
                raise self.finishedErrors(result, job_id, 'promote', 'delayed')
        return None

    def remove(self, job_id: str, remove_children: bool):
        keys = self.getKeys([''])
        args = [job_id, 1 if remove_children else 0]

        return self.commands["removeJob"](keys=keys, args=args)

    def getCounts(self, types):
        keys = self.getKeys([''])
        transformed_types = list(
            map(lambda type: 'wait' if type == 'waiting' else type, types))

        return self.commands["getCounts"](keys=keys, args=transformed_types)

    async def getState(self, job_id: str):
        keys = self.getKeys(['completed', 'failed', 'delayed', 'active', 'wait',
                'paused', 'waiting-children', 'prioritized'])

        args = [job_id, self.toKey(job_id)]

        redis_version = await self.redisConnection.getRedisVersion()

        if isRedisVersionLowerThan(redis_version, '6.0.6'):
            result = await self.commands["getState"](keys=keys, args=args)
            return result

        result = await self.commands["getStateV2"](keys=keys, args=args)
        return result

    async def isJobInList(self, list_key: str, job_id: str):
        redis_version = await self.redisConnection.getRedisVersion()

        result = None
        if isRedisVersionLowerThan(redis_version, '6.0.6'):
            result = await self.commands["getState"](keys=[list_key], args=[job_id])
        else:
            result = await self.redisClient.lpos(list_key, job_id)
        return isinstance(result, int)

    async def changePriority(self, job_id: str, priority:int = 0, lifo:bool = False):
        keys = [self.keys['wait'],
            self.keys['paused'],
            self.keys['meta'],
            self.keys['prioritized'],
            self.keys['pc'],
            self.keys['marker']]
        
        args = [priority, self.toKey(job_id), job_id, 1 if lifo else 0]

        result = await self.commands["changePriority"](keys=keys, args=args)

        if result is not None:
            if result < 0:
                raise self.finishedErrors(result, job_id, 'changePriority', None)
        return None

    async def updateData(self, job_id: str, data):
        keys = [self.toKey(job_id)]
        data_json = json.dumps(data, separators=(',', ':'))
        args = [data_json]

        result = await self.commands["updateData"](keys=keys, args=args)

        if result is not None:
            if result < 0:
                raise self.finishedErrors(result, job_id, 'updateData', None)
        return None

    async def reprocessJob(self, job: Job, state: str):
        keys = [self.toKey(job.id)]
        keys.append(self.keys['events'])
        keys.append(self.keys[state])
        keys.append(self.keys['wait'])
        keys.append(self.keys['meta'])
        keys.append(self.keys['paused'])
        keys.append(self.keys['marker'])

        args = [
            job.id,
            ("R" if job.opts.get("lifo") else "L") + "PUSH",
            "failedReason" if state == "failed" else "returnvalue",
            state
            ]

        result = await self.commands["reprocessJob"](keys=keys, args=args)

        if result is not None:
            if result < 0:
                raise self.finishedErrors(result, job.id, 'reprocessJob', state)
        return None

    def pause(self, pause: bool = True):
        """
        Pause or resume a queue
        """
        src = "wait" if pause else "paused"
        dst = "paused" if pause else "wait"
        keys = self.getKeys([src, dst, 'meta', 'prioritized', 'events', 'delayed', 'marker'])
        return self.commands["pause"](keys, args=["paused" if pause else "resumed"])

    async def obliterate(self, count: int, force: bool = False):
        """
        Remove a queue completely
        """
        keys = self.getKeys(['meta', ''])
        result = await self.commands["obliterate"](keys, args=[count, force or ""])
        if (result < 0):
            if (result == -1):
                raise Exception("Cannot obliterate non-paused queue")
            if (result == -2):
                raise Exception("Cannot obliterate queue with active jobs")
        return result

    def moveJobsToWaitArgs(self, state: str, count: int, timestamp: int) -> int:
        keys = self.getKeys(
            ['', 'events', state, 'wait', 'paused', 'meta', 'marker'])
        
        args = [count or 1000, timestamp or round(time.time()*1000), state]
        return (keys, args)

    async def retryJobs(self, state: str, count: int, timestamp: int):
        """
        Retry jobs that are in failed or completed state
        """
        current_state = state or 'failed'
        keys, args = self.moveJobsToWaitArgs(current_state, count, timestamp)

        result = await self.commands["moveJobsToWait"](keys=keys, args=args)
        return result

    async def promoteJobs(self, count: int):
        """
        Promote jobs in delayed state
        """
        keys, args = self.moveJobsToWaitArgs('delayed', count, 1.7976931348623157e+308)

        result = await self.commands["moveJobsToWait"](keys=keys, args=args)
        return result

    async def moveToActive(self, token: str, opts: dict) -> list[Any]:
        """
        Add an item to the queue
        """
        timestamp = round(time.time() * 1000)
        lockDuration = opts.get("lockDuration", 0)
        limiter = opts.get("limiter", None)

        keys = self.getKeys(['wait', 'active', 'prioritized', 'events',
                            'stalled', 'limiter', 'delayed', 'paused', 'meta', 'pc', 'marker'])
        packedOpts = msgpack.packb(
            {"token": token, "lockDuration": lockDuration, "limiter": limiter}, use_bin_type=True)
        args = [self.keys[''], timestamp, packedOpts]

        result = await self.commands["moveToActive"](keys=keys, args=args)

        return raw2NextJobData(result)

    def moveToCompleted(self, job: Job, val: Any, removeOnComplete, token: str, opts: dict, fetchNext=True):
        return self.moveToFinished(job, val, "returnvalue", removeOnComplete, "completed", token, opts, fetchNext)

    def moveToFailed(self, job: Job, failedReason: str, removeOnFailed, token: str, opts: dict, fetchNext=True):
        return self.moveToFinished(job, failedReason, "failedReason", removeOnFailed, "failed", token, opts, fetchNext)

    async def updateProgress(self, job_id: str, progress):
        keys = [self.toKey(job_id), self.keys['events'], self.keys['meta']]
        progress_json = json.dumps(progress, separators=(',', ':'))
        args = [job_id, progress_json]
        result = await self.commands["updateProgress"](keys=keys, args=args)

        if result is not None:
            if result < 0:
                raise self.finishedErrors(result, job_id, 'updateProgress', None)
        return None

    def moveToFinishedArgs(self, job: Job, val: Any, propVal: str, shouldRemove, target, token: str, opts: dict, fetchNext=True) -> list[Any] | None:
        transformed_value = json.dumps(val, separators=(',', ':'))
        timestamp = round(time.time() * 1000)
        metricsKey = self.toKey('metrics:' + target)

        keys = self.getKeys(['wait', 'active', 'prioritized', 'events',
                            'stalled', 'limiter', 'delayed', 'paused', 'meta', 'pc', target])
        keys.append(self.toKey(job.id))
        keys.append(metricsKey)
        keys.append(self.keys['marker'])

        def getKeepJobs(shouldRemove: bool | dict | int | None):
            if type(shouldRemove) == int:
                return {"count": shouldRemove}

            if type(shouldRemove) == dict:
                return shouldRemove

            if shouldRemove:
                return {"count": 0}

            if not shouldRemove or shouldRemove is None:
                return {"count": -1}

        def getMetricsSize(opts: dict):
            metrics = opts.get("metrics")
            if metrics is not None:
                return metrics.get("maxDataPoints", "")
            return ""

        def getFailParentOnFailure(job: Job):
            opts = job.opts
            if opts is not None:
                return opts.get("failParentOnFailure", False)

        keepJobs = getKeepJobs(shouldRemove)

        packedOpts = msgpack.packb({
            "token": token,
            "keepJobs": keepJobs,
            "limiter": opts.get("limiter"),
            "lockDuration": opts.get("lockDuration"),
            "attempts": job.attempts,
            "attemptsMade": job.attemptsMade,
            "maxMetricsSize": getMetricsSize(opts),
            "fpof": getFailParentOnFailure(job),
        }, use_bin_type=True)

        args = [job.id, timestamp, propVal, transformed_value or "", target,
                fetchNext and "1" or "", self.keys[''], packedOpts]
        return (keys, args)

    def moveToFailedArgs(self, job: Job, failed_reason: str, shouldRemove, token: str, opts: dict, fetchNext=True):
        return self.moveToFinishedArgs(
            job, failed_reason, 'failedReason', shouldRemove, 'failed',
            token, opts, fetchNext
        )

    async def moveToFinished(self, job: Job, val: Any, propVal: str, shouldRemove, target, token: str, opts: dict, fetchNext=True) -> list[Any] | None:
        keys, args = self.moveToFinishedArgs(job, val, propVal, shouldRemove, target, token, opts, fetchNext)

        result = await self.commands["moveToFinished"](keys=keys, args=args)

        if result is not None:
            if type(result) == int and result < 0:
                raise self.finishedErrors(result, job.id, 'finished', 'active')
            return raw2NextJobData(result)
        return None

    def extendLock(self, jobId: str, token: str, duration: int, client: Redis = None):
        keys = [self.toKey(jobId) + ":lock", self.keys['stalled']]
        args = [token, duration, jobId]
        return self.commands["extendLock"](keys, args, client)

    def moveStalledJobsToWait(self, maxStalledCount: int, stalledInterval: int):
        keys = self.getKeys(['stalled', 'wait', 'active', 'failed',
                            'stalled-check', 'meta', 'paused', 'marker', 'events'])
        args = [maxStalledCount, self.keys[''], round(
            time.time() * 1000), stalledInterval]
        return self.commands["moveStalledJobsToWait"](keys, args)

    def finishedErrors(self, code: int, jobId: str, command: str, state: str) -> TypeError:
        if code == ErrorCode.JobNotExist.value:
            return TypeError(f"Missing key for job {jobId}.{command}")
        elif code == ErrorCode.JobLockNotExist.value:
            return TypeError(f"Missing lock for job {jobId}.{command}")
        elif code == ErrorCode.JobNotInState.value:
            return TypeError(f"Job {jobId} is not in the state {state}.{command}")
        elif code == ErrorCode.JobPendingDependencies.value:
            return TypeError(f"Job {jobId} has pending dependencies.{command}")
        elif code == ErrorCode.ParentJobNotExist.value:
            return TypeError(f"Missing key for parent job {jobId}.{command}")
        elif code == ErrorCode.JobLockMismatch.value:
            return TypeError(f"Lock mismatch for job {jobId}. Cmd {command} from {state}")
        elif code == ErrorCode.ParentJobCannotBeReplaced.value:
            return TypeError(f"The parent job {jobId} cannot be replaced. {command}")
        else:
            return TypeError(f"Unknown code {str(code)} error for {jobId}.{command}")


def raw2NextJobData(raw: list[Any]) -> list[Any] | None:
    if raw:
        result = [None, raw[1], None, None] if len(raw) == 2 else [None, raw[1], raw[2], raw[3]]
        if raw[0]:
            result[0]= array2obj(raw[0])
        return result
    return [None, None, None, None]


def array2obj(arr: list[str]) -> dict[str, str]:
    obj = {}
    for i in range(0, len(arr), 2):
        obj[arr[i]] = arr[i + 1]
    return obj


def convert_to_int(text: str):
    try:
        result = int(text)
        return result
    except ValueError:
        return 0



from typing import Callable
from uuid import uuid4
from bullmq.custom_errors import WaitingChildrenError
from bullmq.scripts import Scripts
from bullmq.redis_connection import RedisConnection
from bullmq.event_emitter import EventEmitter
from bullmq.job import Job
from bullmq.timer import Timer
from bullmq.types import WorkerOptions
from bullmq.utils import extract_result

import asyncio
import traceback
import time
import math

maximum_block_timeout = 10
# 1 millisecond is chosen because the granularity of our timestamps are milliseconds.
# Obviously we can still process much faster than 1 job per millisecond but delays and rate limits will never work with more accuracy than 1ms.
minimum_block_timeout = 0.001

class Worker(EventEmitter):
    def __init__(self, name: str, processor: Callable[[Job, str], asyncio.Future], opts: WorkerOptions = {}):
        super().__init__()
        self.name = name
        self.processor = processor
        final_opts = {
            "concurrency": 1,
            "lockDuration": 30000,
            "maxStalledCount": 1,
            "stalledInterval": 30000,
        }
        final_opts.update(opts or {})
        self.opts = final_opts
        redis_opts = opts.get("connection", {})
        self.redisConnection = RedisConnection(redis_opts)
        self.blockingRedisConnection = RedisConnection(redis_opts)
        self.client = self.redisConnection.conn
        self.bclient = self.blockingRedisConnection.conn
        self.prefix = opts.get("prefix", "bull")
        self.scripts = Scripts(opts.get("prefix", "bull"), name, self.redisConnection)
        self.closing = False
        self.forceClosing = False
        self.closed = False
        self.running = False
        self.processing = set()
        self.jobs = set()
        self.id = uuid4().hex
        self.waiting = None
        self.blockUntil = 0
        self.limitUntil = 0
        self.drained = False
        self.qualifiedName = self.scripts.queue_keys.getQueueQualifiedName(name)

        if opts.get("autorun", True):
            asyncio.ensure_future(self.run())

    async def run(self):
        if self.running:
            raise Exception("Worker is already running")

        self.timer = Timer(
            (self.opts.get("lockDuration") / 2) / 1000, self.extendLocks)
        self.stalledCheckTimer = Timer(self.opts.get(
            "stalledInterval") / 1000, self.runStalledJobsCheck)
        self.running = True
        jobs = []

        token_postfix = 0

        while not self.closed:
            while not self.waiting and len(self.processing) < self.opts.get("concurrency") and not self.closing:
                token_postfix+=1
                token = f'{self.id}:{token_postfix}'
                waiting_job = asyncio.ensure_future(self.getNextJob(token))
                self.processing.add(waiting_job)

            try:
                jobs, pending = await getCompleted(self.processing)

                jobs_to_process = [self.processJob(job, job.token) for job in jobs]
                processing_jobs = [asyncio.ensure_future(
                    j) for j in jobs_to_process]
                pending.update(processing_jobs)
                self.processing = pending

                if (len(jobs) == 0 or len(self.processing) == 0) and self.closing:
                    # We are done processing so we can close the queue
                    break

            except Exception as e:
                # This should never happen or we will have an endless loop
                print("ERROR:", e)
                traceback.print_exc()
                return

        self.running = False
        self.timer.stop()
        self.stalledCheckTimer.stop()

    async def getNextJob(self, token: str):
        """
        Returns a promise that resolves to the next job in queue.
        @param token: worker token to be assigned to retrieved job
        @returns a Job or undefined if no job was available in the queue.
        """
        if not self.waiting and self.drained:
            self.waiting = self.waitForJob()

            try:
                self.blockUntil = await self.waiting
                timestamp = int(time.time() * 1000)

                if self.blockUntil <= 0 or self.blockUntil <= timestamp:
                    job_instance = await self.moveToActive(token)
                    return job_instance
            finally:
                self.waiting = None
        else:
            job_instance = await self.moveToActive(token)
            return job_instance

    async def moveToActive(self, token: str):
        result = await self.scripts.moveToActive(token, self.opts)
        job_data = None
        id = None
        limit_until = None
        delay_until = None

        if result:
            job_data, id, limit_until, delay_until = result

        return self.nextJobFromJobData(job_data, id, limit_until, delay_until, token)

    def nextJobFromJobData(self, job_data = None, job_id: str = None, limit_until: int = 0,
        delay_until: int = 0, token: str = None):
        self.limitUntil = max(limit_until, 0) or 0

        if not job_data:
            if not self.drained:
                self.drained = True
                self.blockUntil = 0

        if delay_until:
            self.blockUntil = max(delay_until, 0) or 0

        if job_data:
            self.drained = False
            job_instance = Job.fromJSON(self, job_data, job_id)
            job_instance.token = token
            return job_instance

    async def waitForJob(self):
        block_timeout = self.getBlockTimeout(self.blockUntil)
        block_timeout = block_timeout if self.blockingRedisConnection.capabilities.get("canDoubleTimeout", False) else math.ceil(block_timeout)
        block_timeout = min(block_timeout, maximum_block_timeout)

        result = await self.bclient.bzpopmin(self.scripts.keys["marker"], block_timeout)
        if result:
            [_key, member, score] = result

            if member:
                return int(score)
            else:
                return 0
        return 0

    def getBlockTimeout(self, block_until: int):
        if block_until:
            block_delay = block_until - int(time.time() * 1000)
            if block_delay < 1:
                return minimum_block_timeout
            else:
                return block_delay / 1000
        else:
            return max(self.opts.get("drainDelay", 5), minimum_block_timeout)

    async def processJob(self, job: Job, token: str):
        try:
            self.jobs.add((job, token))
            result = await self.processor(job, token)
            if not self.forceClosing:
                # Currently we do not support pre-fetching jobs as in NodeJS version.
                # nextJob = await self.scripts.moveToCompleted(job, result, job.opts.get("removeOnComplete", False), token, self.opts, fetchNext=not self.closing)
                await self.scripts.moveToCompleted(job, result, job.opts.get("removeOnComplete", False), token, self.opts, fetchNext=False)
                job.returnvalue = result
                job.attemptsMade = job.attemptsMade + 1
            self.emit("completed", job, result)
        except WaitingChildrenError:
            return
        except Exception as err:
            try:
                print("Error processing job", err)
                if not self.forceClosing:
                    await job.moveToFailed(err, token)

                self.emit("failed", job, err)
            except Exception as err:
                print("Error moving job to failed", err)
                self.emit("error", err, job)
        finally:
            self.jobs.remove((job, token))

    async def extendLocks(self):
        # Renew all the locks for the jobs that are still active
        try:
            multi = self.client.pipeline()
            for job, token in self.jobs:
                await self.scripts.extendLock(job.id, token, self.opts.get("lockDuration"), multi)
            result = await multi.execute()

            # result includes an object with locks that may not have been renewed.
            # We should emit an error for each of those jobs.
            #    for jobId, err in result.items():
            #    self.emit("error", "could not renew lock for job " + jobId)

        except Exception as e:
            print("Error renewing locks", e)
            traceback.print_exc()

    async def runStalledJobsCheck(self):
        try:
            failed, stalled = await self.scripts.moveStalledJobsToWait(self.opts.get("maxStalledCount"), self.opts.get("stalledInterval"))
            for jobId in failed:
                self.emit("failed", jobId,
                          "job stalled more than allowable limit")
            for jobId in stalled:
                self.emit("stalled", jobId)

        except Exception as e:
            print("Error checking stalled jobs", e)
            self.emit('error', e)

    async def close(self, force: bool = False):
        """
        Close the worker
        """
        self.closing = True
        if force:
            self.forceClosing = True
            self.cancelProcessing()

        await self.blockingRedisConnection.close()
        await self.redisConnection.close()

    def cancelProcessing(self):
        for job in self.processing:
            if not job.done():
                job.cancel()


async def getCompleted(task_set: set) -> tuple[list[Job], set]:
    job_set, pending = await asyncio.wait(task_set, return_when=asyncio.FIRST_COMPLETED)
    jobs = [extract_result(job_task) for job_task in job_set]
    # we filter `None` out to remove:
    # a) an empty 'completed jobs' list; and
    # b) a failed extract_result
    jobs = list(filter(lambda j: j is not None, jobs))
    return jobs, pending



import redis.asyncio as redis
from redis.backoff import ExponentialBackoff
from redis.retry import Retry
from redis.exceptions import (
   BusyLoadingError,
   ConnectionError,
   TimeoutError
)
import warnings
from bullmq.utils import isRedisVersionLowerThan

class RedisConnection:
    """
    RedisConnection class
    """

    minimum_version = '5.0.0'
    recommended_minimum_version = '6.2.0'

    capabilities = {
        "canDoubleTimeout": False
    }

    def __init__(self, redisOpts: dict | str = {}):
        self.version = None
        retry = Retry(ExponentialBackoff(), 3)
        retry_errors = [BusyLoadingError, ConnectionError, TimeoutError]

        if isinstance(redisOpts, dict):
            defaultOpts = {
                "host": "localhost",
                "port": 6379,
                "db": 0,
                "password": None,
                "username": None,
            }
            finalOpts = {**defaultOpts, **redisOpts}

            self.conn = redis.Redis(decode_responses=True, retry=retry, retry_on_error=retry_errors, **finalOpts)
        else:
            self.conn = redis.from_url(redisOpts, decode_responses=True, retry=retry,
                retry_on_error=retry_errors)

    def disconnect(self):
        """
        Disconnect from Redis
        """
        return self.conn.disconnect()

    def close(self):
        """
        Close the connection
        """
        return self.conn.close()

    async def getRedisVersion(self):
        if self.version is not None:
            return self.version

        doc = await self.conn.info()
        if doc.get("maxmemory_policy") != "noeviction":
            warnings.warn(f'IMPORTANT! Eviction policy is {doc.get("maxmemory_policy")}. It should be "noeviction"')

        self.version = doc.get("redis_version")

        self.capabilities = {
            "canDoubleTimeout": not isRedisVersionLowerThan(self.version, '6.0.0')
        }
        return self.version



from __future__ import annotations
from typing import List, Any, TYPE_CHECKING
from bullmq.scripts import Scripts
from bullmq.backoffs import Backoffs
if TYPE_CHECKING:
    from bullmq.queue import Queue
from bullmq.types import JobOptions
from bullmq.utils import get_parent_key

import json
import time
import traceback


optsDecodeMap = {
    'fpof': 'failParentOnFailure',
    'kl': 'keepLogs',
}

optsEncodeMap = {v: k for k, v in optsDecodeMap.items()}


class Job:
    """
    This class represents a Job in the queue. Normally job are implicitly created when
    you add a job to the queue with methods such as Queue.addJob( ... )

    A Job instance is also passed to the Worker's process function.
    """

    def __init__(self, queue: Queue, name: str, data: Any, opts: JobOptions = {}, job_id: str = None):
        self.name = name
        self.id = opts.get("jobId", None) or job_id
        self.progress = 0
        self.timestamp = opts.get("timestamp", round(time.time() * 1000))
        final_opts = {"attempts": 0, "delay": 0}
        final_opts.update(opts or {})
        final_opts.update({"backoff": Backoffs.normalize(opts.get('backoff'))})
        self.discarded = False
        self.opts = final_opts
        self.queue = queue
        self.delay = opts.get("delay", 0)
        self.attempts = opts.get("attempts", 1)
        self.attemptsMade = 0
        self.attemptsStarted = 0
        self.data = data
        self.removeOnComplete = opts.get("removeOnComplete", True)
        self.removeOnFail = opts.get("removeOnFail", False)
        self.processedOn = 0
        self.finishedOn = 0
        self.returnvalue = None
        self.failedReason = None
        self.repeatJobKey = None
        self.token: str = None
        parent = opts.get("parent")
        self.parentKey = get_parent_key(parent)
        self.parent = {"id": parent.get("id"), "queueKey": parent.get("queue")} if parent else None
        self.stacktrace: List[str] = []
        self.scripts = Scripts(queue.prefix, queue.name, queue.redisConnection)
        self.queueQualifiedName = queue.qualifiedName

    def updateData(self, data):
        self.data = data
        return self.scripts.updateData(self.id, data)

    async def promote(self):
        await self.scripts.promote(self.id)
        self.delay = 0

    def retry(self, state: str = "failed"):
        self.failedReason = None
        self.finishedOn = None
        self.processedOn = None
        self.returnvalue = None
        return self.scripts.reprocessJob(self, state)

    def getState(self):
        return self.scripts.getState(self.id)

    def changePriority(self, opts: dict):
        return self.scripts.changePriority(self.id, opts.get("priority", 0), opts.get("lifo", False))

    def updateProgress(self, progress):
        self.progress = progress
        return self.scripts.updateProgress(self.id, progress)

    async def remove(self, opts: dict = {}):
        removed = await self.scripts.remove(self.id, opts.get("removeChildren", True))

        if not removed:
            raise Exception(f"Job {self.id} could not be removed because it is locked by another worker")

    def isCompleted(self):
        """
        Returns true if the job has completed.
        """
        return self.isInZSet('completed')

    def isFailed(self):
        """
        Returns true if the job has failed.
        """
        return self.isInZSet('failed')

    def isDelayed(self):
        """
        Returns true if the job is delayed.
        """
        return self.isInZSet('delayed')

    def isWaitingChildren(self):
        """
        Returns true if the job is waiting for children.
        """
        return self.isInZSet('waiting-children')

    def isActive(self):
        """
        Returns true if the job is active.
        """
        return self.isInList('active')

    async def isWaiting(self):
        return ( await self.isInList('wait') or await self.isInList('paused'))

    async def isInZSet(self, set: str):
        score = await self.queue.client.zscore(self.scripts.toKey(set), self.id)

        return score is not None

    def isInList(self, list_name: str):
        return self.scripts.isJobInList(self.scripts.toKey(list_name), self.id)

    async def moveToFailed(self, err, token:str, fetchNext:bool = False):
        error_message = str(err)
        self.failedReason = error_message

        move_to_failed = False
        finished_on = 0
        delay = 0
        command = 'moveToFailed'

        async with self.queue.redisConnection.conn.pipeline(transaction=True) as pipe:
            await self.saveStacktrace(pipe, error_message)
            if (self.attemptsMade + 1) < self.opts.get('attempts') and not self.discarded:
                delay = await Backoffs.calculate(
                    self.opts.get('backoff'), self.attemptsMade + 1,
                    err, self, self.queue.opts.get("settings") and self.queue.opts['settings'].get("backoffStrategy")
                    )
                if delay == -1:
                    move_to_failed = True
                elif delay:
                    keys, args = self.scripts.moveToDelayedArgs(
                        self.id,
                        round(time.time() * 1000) + delay,
                        token,
                        delay
                    )

                    await self.scripts.commands["moveToDelayed"](keys=keys, args=args, client=pipe)
                    command = 'delayed'
                else:
                    keys, args = self.scripts.retryJobArgs(self.id, self.opts.get("lifo", False), token)

                    await self.scripts.commands["retryJob"](keys=keys, args=args, client=pipe)
                    command = 'retryJob'
            else:
                move_to_failed = True

            if move_to_failed:
                keys, args = self.scripts.moveToFailedArgs(
                    self, error_message, self.opts.get("removeOnFail", False),
                    token, self.opts, fetchNext
                )
                await self.scripts.commands["moveToFinished"](keys=keys, args=args, client=pipe)
                finished_on = args[1]

            results = await pipe.execute()
            code = results[1]

            if code < 0:
                raise self.scripts.finishedErrors(code, self.id, command, 'active')

        if finished_on and type(finished_on) == int:
            self.finishedOn = finished_on

        if delay and type(delay) == int:
            self.delay = delay

        self.attemptsMade = self.attemptsMade + 1

    def log(self, logRow: str):
        return Job.addJobLog(self.queue, self.id, logRow, self.opts.get("keepLogs", 0))

    async def saveStacktrace(self, pipe, err:str):
        stacktrace = traceback.format_exc()
        stackTraceLimit = self.opts.get("stackTraceLimit")

        if stacktrace:
            self.stacktrace.append(stacktrace)
            if self.opts.get("stackTraceLimit"):
                self.stacktrace = self.stacktrace[-(stackTraceLimit-1):stackTraceLimit]

        keys, args = self.scripts.saveStacktraceArgs(
            self.id, json.dumps(self.stacktrace, separators=(',', ':')), err)

        await self.scripts.commands["saveStacktrace"](keys=keys, args=args, client=pipe)

    def moveToWaitingChildren(self, token, opts:dict):
        return self.scripts.moveToWaitingChildren(self.id, token, opts)

    @staticmethod
    def fromJSON(queue: Queue, rawData: dict, jobId: str | None = None):
        """
        Instantiates a Job from a JobJsonRaw object (coming from a deserialized JSON object)

        @param queue: the queue where the job belongs to.
        @param json: the plain object containing the job.
        @param jobId: an optional job id (overrides the id coming from the JSON object)
        """
        data = json.loads(rawData.get("data", '{}'))
        opts = optsFromJSON(json.loads(rawData.get("opts", '{}')))

        job = Job(queue, rawData.get("name"), data, opts)
        job.id = jobId or rawData.get("id", b'').decode("utf-8")

        job.progress = json.loads(rawData.get("progress",  '0'))
        job.delay = int(rawData.get("delay", "0"))
        job.timestamp = int(rawData.get("timestamp", "0"))

        if rawData.get("finishedOn"):
            job.finishedOn = int(rawData.get("finishedOn"))

        if rawData.get("processedOn"):
            job.processedOn = int(rawData.get("processedOn"))

        if rawData.get("rjk"):
            job.repeatJobKey = rawData.get("rjk")

        if rawData.get("ats"):
            job.attemptsStarted = int(rawData.get("ats"))

        job.failedReason = rawData.get("failedReason")
        job.attemptsMade = int(rawData.get("attemptsMade") or rawData.get("atm") or "0")

        returnvalue = rawData.get("returnvalue")
        if type(returnvalue) == str:
            job.returnvalue = getReturnValue(returnvalue)

        job.stacktrace = json.loads(rawData.get("stacktrace", "[]"))

        if rawData.get("parentKey"):
            job.parentKey = rawData.get("parentKey")

        if rawData.get("parent"):
           job.parent = json.loads(rawData.get("parent"))

        return job

    @staticmethod
    async def fromId(queue: Queue, jobId: str):
        key = f"{queue.prefix}:{queue.name}:{jobId}"
        raw_data = await queue.client.hgetall(key)
        if len(raw_data):
            return Job.fromJSON(queue, raw_data, jobId)

    @staticmethod
    async def addJobLog(queue: Queue, jobId: str, logRow: str, keepLogs: int = 0):
        logs_key = f"{queue.prefix}:{queue.name}:{jobId}:logs"
        multi = await queue.client.pipeline()

        multi.rpush(logs_key, logRow)

        if keepLogs:
            multi.ltrim(logs_key, -keepLogs, -1)
        
        result = await multi.execute()

        return min(keepLogs, result[0]) if keepLogs else result[0]

def optsFromJSON(rawOpts: dict) -> dict:
    # opts = json.loads(rawOpts)
    opts = rawOpts

    option_entries = opts.items()

    options = {}
    for item in option_entries:
        attribute_name = item[0]
        value = item[1]
        if attribute_name in optsDecodeMap:
            options[optsDecodeMap[attribute_name]] = value
        else:
            options[attribute_name] = value

    return options


def getReturnValue(value: Any):
    try:
        return json.loads(value)
    except Exception as err:
        return value



from enum import Enum


class ErrorCode(Enum):
    JobNotExist = -1
    JobLockNotExist = -2
    JobNotInState = -3
    JobPendingDependencies = -4
    ParentJobNotExist = -5
    JobLockMismatch = -6
    ParentJobCannotBeReplaced = -7
